//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include "Segments.h"

ClassImp(Segments)


/////////////////////////////////////////////////////////////////////////
Segments::Segments(void){ 
/////////////////////////////////////////////////////////////////////////
  // zero-out everything
  cout.precision(5);
  cerr.precision(5);
  cout.flags ( ios::fixed );
  cerr.flags ( ios::fixed );
  status=true;
  seg_start.clear();
  seg_end.clear();
  seg_tag.clear();
  time_mark.clear();
  n_mark.clear();
  livetime_tot=0;
}

/////////////////////////////////////////////////////////////////////////
Segments::Segments(const string segfile, const int ncol, const int default_tag){ 
/////////////////////////////////////////////////////////////////////////

  cout.precision(5);
  cerr.precision(5);
  cout.flags ( ios::fixed );
  cerr.flags ( ios::fixed );
  status=true;
  seg_tag.clear();
  livetime_tot=0;

  // read segment file
  if(ncol<0){
    ReadAscii *readseg = new ReadAscii(segfile, "UNKNOWN");
    if(readseg->GetNCol()==2){
      if(!readseg->SetFormat(0,'d')) status=false;
      if(!readseg->SetFormat(1,'d')) status=false;
      status=readseg->GetCol(seg_start, 0);
      status=readseg->GetCol(seg_end,   1);
      for(int s=0; s<(int)seg_start.size(); s++) seg_tag.push_back(TMath::Max(-1,default_tag));
    }
    else if(readseg->GetNCol()>2&&readseg->GetNCol()<5){
      if(!readseg->SetFormat(1,'d')) status=false;
      if(!readseg->SetFormat(2,'d')) status=false;
      status=readseg->GetCol(seg_start, 1);
      status=readseg->GetCol(seg_end,   2);
      for(int s=0; s<(int)seg_start.size(); s++) seg_tag.push_back(TMath::Max(-1,default_tag));
    }
    else if(readseg->GetNCol()==5){
      if(!readseg->SetFormat(1,'d')) status=false;
      if(!readseg->SetFormat(2,'d')) status=false;
      if(!readseg->SetFormat(4,'i')) status=false;
      status=readseg->GetCol(seg_start, 1);
      status=readseg->GetCol(seg_end,   2);
      status=readseg->GetCol(seg_tag,   4);
      if(default_tag>=0){// tag selection
	for(int s=0; s<(int)seg_start.size(); s++){
	  if(seg_tag[s]!=default_tag){
	    seg_start.erase(seg_start.begin()+s);
	    seg_end.erase(seg_end.begin()+s);
	    seg_tag.erase(seg_tag.begin()+s);
	  }
	}
      }
    }
    else{
      cerr<<"Segments::Segments: number of col = "<<readseg->GetNCol()<<" is not supported"<<endl;
      status=false;
    }
    delete readseg;
  }
  else if(ncol==2){
    ReadAscii *readseg = new ReadAscii(segfile, "d;d");
    if(readseg->GetNRow()){// empty file
      status =readseg->GetCol(seg_start, 0);
      status*=readseg->GetCol(seg_end,   1);
      if(status) for(int s=0; s<(int)seg_start.size(); s++) seg_tag.push_back(TMath::Max(-1,default_tag));
    }
    delete readseg;
  }
  else if(ncol==3){
    ReadAscii *readseg = new ReadAscii(segfile, "d;d;d");
    if(readseg->GetNRow()){// empty file
      status =readseg->GetCol(seg_start, 1);
      status*=readseg->GetCol(seg_end,   2);
      if(status) for(int s=0; s<(int)seg_start.size(); s++) seg_tag.push_back(TMath::Max(-1,default_tag));
    }
    delete readseg;
  }
  else if(ncol==4){
    ReadAscii *readseg = new ReadAscii(segfile, "d;d;d;d");
    if(readseg->GetNRow()){// empty file
      status =readseg->GetCol(seg_start, 1);
      status*=readseg->GetCol(seg_end,   2);
      if(status) for(int s=0; s<(int)seg_start.size(); s++) seg_tag.push_back(TMath::Max(-1,default_tag));
    }
    delete readseg;
  }
  else if(ncol==5){
    ReadAscii *readseg = new ReadAscii(segfile, "d;d;d;d;i");
    if(readseg->GetNRow()){// empty file
      status =readseg->GetCol(seg_start, 1);
      status*=readseg->GetCol(seg_end,   2);
      status*=readseg->GetCol(seg_tag,   4);// FIXME: no check against tag<-1
    }
    delete readseg;
  }
  else{
    cerr<<"Segments::Segments: number of col = "<<ncol<<" is not supported"<<endl;
    status=false;
  }

  // check input consistency
  if(status) status*=CheckConsistency();

  // merge overlapping segments
  if(status) status*=MergeOverlaps();

  // set time marks
  if(status) status*=SetMarks();

  // compute livetime
  if(status) status*=ComputeLiveTime();
  
  if(!status) cerr<<"Segments::Segments: "<<segfile<<" initialization failed"<<endl;
}

/////////////////////////////////////////////////////////////////////////
Segments::Segments(TTree *segment_tree){ 
/////////////////////////////////////////////////////////////////////////

  cout.precision(5);
  cerr.precision(5);
  cout.flags ( ios::fixed );
  cerr.flags ( ios::fixed );
  status=true;
  livetime_tot=0;
  seg_start.clear();
  seg_end.clear();
  seg_tag.clear();

  // read tree
  double sstart, send;
  int stag=-1;
  if(segment_tree->SetBranchAddress("start",&sstart)<0||segment_tree->SetBranchAddress("end",&send)<0){
    cerr<<"Segments::Segments: input segment tree does not have the correct format"<<endl;
    status=false;
  }
  if(segment_tree->GetBranchStatus("tag")) segment_tree->SetBranchAddress("tag", &stag);


  // get segments
  if(status){
    for(int s=0; s<segment_tree->GetEntries(); s++){
      segment_tree->GetEntry(s);
      seg_start.push_back(sstart);
      seg_end.push_back(send);
      seg_tag.push_back(TMath::Max(-1,stag));
    }
  }

  // check input consistency
  if(status) status*=CheckConsistency();

  // merge overlapping segments
  if(status) status*=MergeOverlaps();

  // set time marks
  if(status) status*=SetMarks();

  // compute livetime
  if(status) status*=ComputeLiveTime();
  
  if(!status) cerr<<"Segments::Segments: initialization failed"<<endl;
}

/////////////////////////////////////////////////////////////////////////
Segments::Segments(const vector<double> segment_starts, const vector<double> segment_ends){ 
/////////////////////////////////////////////////////////////////////////

  // parameters
  seg_start=segment_starts;
  seg_end=segment_ends;
  
  // init
  cout.precision(5);
  cerr.precision(5);
  cout.flags ( ios::fixed );
  cerr.flags ( ios::fixed );
  status=true;
  livetime_tot=0;
  for(int s=0; s<(int)seg_start.size(); s++) seg_tag.push_back(-1);

  // check input consistency
  if(status) status*=CheckConsistency();

  // merge overlapping segments
  if(status) status*=MergeOverlaps();

  // set time marks
  if(status) status*=SetMarks();

  // compute livetime
  if(status) status*=ComputeLiveTime();
  
  if(!status) cerr<<"Segments::Segments: initialization failed"<<endl;
}

/////////////////////////////////////////////////////////////////////////
Segments::Segments(const double segment_start, const double segment_end, const int segment_tag){ 
/////////////////////////////////////////////////////////////////////////

  // parameters
  seg_start.push_back(segment_start);
  seg_end.push_back(segment_end);
  seg_tag.push_back(TMath::Max(-1,segment_tag));
  
  // init
  cout.precision(5);
  cerr.precision(5);
  cout.flags ( ios::fixed );
  cerr.flags ( ios::fixed );
  status=true;
  livetime_tot=0;

  // check input consistency
  if(status) status*=CheckConsistency();

  // merge overlapping segments
  if(status) status*=MergeOverlaps();

  // set time marks
  if(status) status*=SetMarks();

  // compute livetime
  if(status) status*=ComputeLiveTime();
  
  if(!status) cerr<<"Segments::Segments: initialization failed"<<endl;
}

/////////////////////////////////////////////////////////////////////////
Segments::~Segments(void){
/////////////////////////////////////////////////////////////////////////
  Reset();
}

/////////////////////////////////////////////////////////////////////////
void Segments::Reset(void){
/////////////////////////////////////////////////////////////////////////
  seg_start.clear();
  seg_end.clear();
  seg_tag.clear();
  livetime_tot=0;
  time_mark.clear();
  n_mark.clear();
  status=true;
  return;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::AddSegment(const double gps_start, const double gps_end, const int tag){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;
  if(gps_end<=gps_start){
    cerr<<"Segments::AddSegment: the segment has a negative duration: "<<gps_end-gps_start<<endl;
    return false;
  }
 
  // **** insert first segment
  if(!seg_start.size()){
    seg_start.push_back(gps_start);
    seg_end.push_back(gps_end);
    seg_tag.push_back(TMath::Max(tag,-1));
  }

  // **** just append
  else if(gps_start>=seg_start.back()){
    seg_start.push_back(gps_start);
    seg_end.push_back(gps_end);
    seg_tag.push_back(TMath::Max(tag,-1));
    status*=MergeOverlaps(seg_start.size()-1);
  }

  //**** or insert at the right place
  else{
    // get closest segment
    vector <double> seg;
    int sseg = GetClosestSegment(gps_start, seg);
    if(sseg<0){// should never happen
      cerr<<"Segments::AddSegment: GetClosestSegment failed"<<endl;
      return false;
    }
    if(gps_start>=seg[0]){// the new segment should be placed after
      seg_start.insert(seg_start.begin()+sseg+1,gps_start);
      seg_end.insert(seg_end.begin()+sseg+1,gps_end);
      seg_tag.insert(seg_tag.begin()+sseg+1,TMath::Max(tag,-1));
    }
    else{// the new segment should be placed before
      seg_start.insert(seg_start.begin()+sseg,gps_start);
      seg_end.insert(seg_end.begin()+sseg,gps_end);
      seg_tag.insert(seg_tag.begin()+sseg,TMath::Max(tag,-1));
    }
    status*=MergeOverlaps(sseg-1);
  }

  // set new time marks
  status*=SetMarks();

  // compute new livetime
  status*=ComputeLiveTime();

  return status;
}


/////////////////////////////////////////////////////////////////////////
bool Segments::Append(Segments *Seg){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;
  if(Seg==NULL||!Seg->GetStatus()){
    cerr<<"Segments::Append: input segments are corrupted"<<endl;
    return false;
  }
  if(!Seg->GetNsegments()) return true; // nothing to do

  // input segments should be AFTER current segments
  if(seg_start.size()&&Seg->GetStart(0)<seg_start[seg_start.size()-1]){
    cerr<<"Segments::Append: input segments are overlapping current segment object -> cannot Append"<<endl;
    return false;
  }

  int s = (int)seg_start.size();

  // append
  vector <double> st = Seg->GetStarts();
  vector <double> en = Seg->GetEnds(); 
  vector <int> ta = Seg->GetTags();
  seg_start.insert(seg_start.end(),st.begin(),st.end());
  seg_end.insert(seg_end.end(),en.begin(),en.end());
  seg_tag.insert(seg_tag.end(),ta.begin(),ta.end());
  st.clear();
  en.clear();
  ta.clear();

  // check input consistency
  status*=CheckConsistency();

  // merge overlapping segments
  status*=MergeOverlaps(s);

  // set new time marks
  status*=SetMarks();

  // compute new livetime
  status*=ComputeLiveTime();

  return status;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::TruncateAfter(const double gps){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;
  
  // get starting mark
  int n_s=-1;
  for(int t=0; t<(int)n_mark.size(); t++){
    if(gps>=time_mark[t]&&gps<time_mark[t+1]){
      n_s=TMath::Max(n_mark[t]-1,0);
      break;
    }
  }

  // empty segment list --> nothing to do
  if(n_s<0) return true;

  // find last segment to keep
  int slast;
  for(slast=n_s; slast<(int)seg_start.size(); slast++){
    if(seg_end[slast]>gps) break;
  }

  // nothing is above gps
  if(slast==(int)seg_start.size()) return true;

  // gps is between 2 segments
  else if(seg_start[slast]>=gps){
    seg_start.erase(seg_start.begin()+slast,seg_start.end());
    seg_end.erase(seg_end.begin()+slast,seg_end.end());
    seg_tag.erase(seg_tag.begin()+slast,seg_tag.end());
  }

  // gps is inside a segment
  else{
    seg_start.erase(seg_start.begin()+slast+1,seg_start.end());
    seg_end.erase(seg_end.begin()+slast+1,seg_end.end());
    seg_tag.erase(seg_tag.begin()+slast+1,seg_tag.end());
    seg_end.pop_back();
    seg_end.push_back(gps);
  }
    
  // set new time marks
  status*=SetMarks();

  // compute new livetime
  status*=ComputeLiveTime();

  return status;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::CheckConsistency(void){
/////////////////////////////////////////////////////////////////////////

// CHECK GWOLLUM CONVENTION

  // format check
  if(seg_start.size()!=seg_end.size()){
    cerr<<"Segments::CheckConsistency: The segment list has the wrong format"<<endl;
    return false;
  }
  
  // check the segment chronology
  for(int s=0; s<(int)seg_start.size(); s++){
    if(seg_start[s]<700000000){
      cerr<<"Segments::CheckConsistency: segment "<<seg_start[s]<<" "<<seg_end[s]<<" is too old: before Mar 12 2002 20:26:27 UTC (GWOLLUM convention)"<<endl;
      return false;
    }
    if(seg_end[s]<=seg_start[s]){
      cerr<<"Segments::CheckConsistency: segment "<<seg_start[s]<<" "<<seg_end[s]<<" has a negative duration (GWOLLUM convention)"<<endl;
      return false;
    }
    if(s&&seg_start[s-1]>seg_start[s]){
      cerr<<"Segments::CheckConsistency: segments should be time sorted (GWOLLUM convention) -- "<<seg_start[s-1]<<">"<<seg_start[s]<<endl;
      return false;
    }
  }

  return true;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::SetMarks(){
/////////////////////////////////////////////////////////////////////////
  
  // initiate
  n_mark.clear();
  time_mark.clear();
  
  // no segments -> no mark
  if(!seg_start.size()) return true;

  n_mark.push_back(0);
  time_mark.push_back((int)floor(seg_start[0]/100000.0)*100000);
  
  for(int s=1; s<(int)seg_start.size(); s++){
    if(floor(seg_start[s]/100000.0)>floor(seg_start[s-1]/100000.0)){
       time_mark.push_back((int)floor(seg_start[s]/100000.0)*100000);
       n_mark.push_back(s);
    }
  }
  time_mark.push_back(((int)floor(seg_end[(int)seg_start.size()-1]/100000)+1)*100000);
  
  return true;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::ComputeLiveTime(void){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;
  livetime_tot=0;
  for(int s=0; s<(int)seg_start.size(); s++) livetime_tot+=(seg_end[s]-seg_start[s]);
  return true;
}

/////////////////////////////////////////////////////////////////////////
double Segments::GetLiveTime(const double gps_start, const double gps_end){
/////////////////////////////////////////////////////////////////////////
  if(!status) return 0;

  double restricted_livetime=0;

  // get starting mark
  int n_s=-1;
  for(int t=0; t<(int)n_mark.size(); t++){
    if(gps_start>=time_mark[t]&&gps_start<time_mark[t+1]){
      n_s=TMath::Max(n_mark[t]-1,0);
      break;
    }
  }

  // empty segment list
  if(n_s<0) return 0;

  // get intersection
  for(int s=n_s; s<(int)seg_start.size(); s++){

    // out of requested gps segment
    if(seg_end[s]<=gps_start) continue;
    if(seg_start[s]>=gps_end) break;

    // add some livetime
    if(seg_start[s]>=gps_start&&seg_end[s]<=gps_end) restricted_livetime+=(seg_end[s]-seg_start[s]);// completely inside
    else if(seg_start[s]<gps_start&&seg_end[s]>gps_start&&seg_end[s]<=gps_end) restricted_livetime+=(seg_end[s]-gps_start);
    else if(seg_start[s]>=gps_start&&seg_start[s]<gps_end&&seg_end[s]>gps_end) restricted_livetime+=(gps_end-seg_start[s]);
    else return gps_end-gps_start;
    
  }

  return restricted_livetime;
}

/////////////////////////////////////////////////////////////////////////
double Segments::GetLiveTime(const int s){
/////////////////////////////////////////////////////////////////////////
  if(!status) return -1.0;
  if(s<0||s>=(int)seg_start.size()){
    cerr<<"Segments::GetLiveTime: there is no "<<s<<" segment"<<endl;
    return -1;
  }
  return seg_end[s]-seg_start[s];
}

/////////////////////////////////////////////////////////////////////////
double Segments::GetLiveTimeWithTag(const int tag){
/////////////////////////////////////////////////////////////////////////
  if(!status) return 0.0;
  double live=0.0;
  for(int s=0; s<(int)seg_start.size(); s++) if(seg_tag[s]==tag) live+=(seg_end[s]-seg_start[s]);
  return live;
}

/////////////////////////////////////////////////////////////////////////
double Segments::GetStart(const int s){
/////////////////////////////////////////////////////////////////////////
  if(!status) return -1;
  if(s<0||s>=(int)seg_start.size()){
    cerr<<"Segments::GetStart: there is no "<<s<<" segment"<<endl;
    return -1;
  }
  return seg_start[s];
}

/////////////////////////////////////////////////////////////////////////
double Segments::GetEnd(const int s){
/////////////////////////////////////////////////////////////////////////
  if(!status) return -1;
  if(s<0||s>=(int)seg_start.size()){
    cerr<<"Segments::GetEnd: there is no "<<s<<" segment"<<endl;
    return -1;
  }
  return seg_end[s];
}

/////////////////////////////////////////////////////////////////////////
double Segments::GetDuration(const int s){
/////////////////////////////////////////////////////////////////////////
  if(!status) return -1;
  if(s<0||s>=(int)seg_start.size()){
    cerr<<"Segments::GetDuration: there is no "<<s<<" segment"<<endl;
    return -1;
  }
  return seg_end[s]-seg_start[s];
}

/////////////////////////////////////////////////////////////////////////
TH1I* Segments::GetHisto(const double offset){
/////////////////////////////////////////////////////////////////////////
  if(!status) return NULL;
  if(!seg_start.size()) return NULL;
  return GetHisto(seg_start[0],seg_end[(int)seg_start.size()-1],offset);
  return NULL;
}

/////////////////////////////////////////////////////////////////////////
TH1I* Segments::GetHisto(const double gps_start, const double gps_end, const double offset){
/////////////////////////////////////////////////////////////////////////
  if(!status) return NULL;

  TH1I *h_seg = new TH1I("h_seg","Segments",1,gps_start+offset,gps_end+offset);
 
  // get starting mark
  int n_s=-1;
  for(int t=0; t<(int)n_mark.size(); t++){
    if(gps_start>=time_mark[t]&&gps_start<time_mark[t+1]){
      n_s=TMath::Max(n_mark[t]-1,0);
      break;
    }
  }

  // return empty histo
  if(n_s<0) return h_seg;
  
  vector <double> blimits;
  vector <int> bvalue;

  // get segments
  for(int s=n_s; s<(int)seg_start.size(); s++){
    if(seg_end[s]<=gps_start) continue;
    if(seg_start[s]>=gps_end) break;
    blimits.push_back(TMath::Max(seg_start[s],gps_start));
    bvalue.push_back(1);
    blimits.push_back(TMath::Min(seg_end[s],gps_end));
    bvalue.push_back(0);
  }
  
  // return empty histo
  if(!blimits.size()) return h_seg;

  // lower bound
  if(gps_start<blimits[0]){
    blimits.insert(blimits.begin(),gps_start);
    bvalue.insert(bvalue.begin(),0);
  }

  // upper bound
  if(gps_end>blimits[(int)blimits.size()-1])
    blimits.push_back(gps_end);
  else bvalue.pop_back();
  
  // make histo
  delete h_seg;
  int nbins=(int)bvalue.size();
  double *bin = new double [nbins+1];
  for(int b=0; b<=nbins; b++) bin[b]=blimits[b]+offset;
  h_seg = new TH1I("h_seg","Segments",nbins,bin);
  for(int b=1; b<=nbins; b++) h_seg->SetBinContent(b,bvalue[b-1]);
  delete bin;

  blimits.clear();
  bvalue.clear();
  return h_seg;
}

/////////////////////////////////////////////////////////////////////////
TTree* Segments::GetTree(void){
/////////////////////////////////////////////////////////////////////////
  if(!status) return NULL;
  return GetTree(0,1e20);
}

/////////////////////////////////////////////////////////////////////////
TTree* Segments::GetTree(const double gps_start, const double gps_end){
/////////////////////////////////////////////////////////////////////////
  if(!status) return NULL;

  double t_s, t_e;
  int t_t;
  TTree *t_seg = new TTree("segments","segments");
  t_seg->Branch("start",&t_s, "start/D");
  t_seg->Branch("end",  &t_e, "end/D");
  t_seg->Branch("tag",  &t_t, "tag/I");

  // get starting mark
  int n_s=0;
  for(int t=0; t<(int)n_mark.size(); t++){
    if(gps_start>=time_mark[t]&&gps_start<time_mark[t+1]){
      n_s=TMath::Max(n_mark[t]-1,0);
      break;
    }
  }

  // get segments
  for(int s=n_s; s<(int)seg_start.size(); s++){
    if(seg_end[s]<=gps_start) continue;
    if(seg_start[s]>=gps_end) break;
    t_s=TMath::Max(seg_start[s],gps_start);
    t_e=TMath::Min(seg_end[s],gps_end);
    t_t=seg_tag[s];
    t_seg->Fill();
  }
    
  return t_seg;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::GetTimeSeries(bool *timeseries, const int sampling_rate, const int gps_start, const int gps_end){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;
  if(timeseries==NULL) return false;

  // segments for the requested time range
  Segments *tmp = new Segments((double)gps_start,(double)gps_end);
  if(!tmp->GetStatus()){ delete tmp; return false; }
  if(!tmp->Intersect(this)) { delete tmp; return false; }
  if(!tmp->Intersect(this)) { delete tmp; return false; }
  if(!tmp->ApplyPadding(-1.0/(double)sampling_rate,1.0/(double)sampling_rate)) { delete tmp; return false; }
  if(!tmp->Intersect(this)) { delete tmp; return false; }

  // vector size
  // FIXME: check size compatibility
  int tsize = (gps_end-gps_start)*sampling_rate;

  // fill timeseries vector
  double gps;
  for(int i=0; i<tsize; i++){
    timeseries[i]=false;
    gps=gps_start+(double)i/sampling_rate;
    if(IsInsideSegment(gps)) timeseries[i]=true;
  }
  
  delete tmp;
  return true;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::IsInsideSegment(const double gps){
/////////////////////////////////////////////////////////////////////////
  int not_used;
  return IsInsideSegment(gps, not_used);
}

/////////////////////////////////////////////////////////////////////////
bool Segments::IsInsideSegment(const double gps, int &seg_n){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;

  int n_s=-1,n_e=-1;

  // get starting mark
  for(int t=0; t<(int)n_mark.size(); t++){
    if(gps>=time_mark[t]&&gps<time_mark[t+1]){
      n_s=TMath::Max(n_mark[t]-1,0);
      if(t!=(int)n_mark.size()-1) n_e=n_mark[t+1];
      else n_e=(int)seg_start.size();
      break;
    }
  }

  if(n_s<0) return false;

  for(int s=n_s; s<n_e; s++) 
    if(gps>=seg_start[s]&&gps<seg_end[s]){ seg_n=s; return true; }

  return false;
}

/////////////////////////////////////////////////////////////////////////
int Segments::GetClosestSegment(const double gps, vector<double> &seg){
/////////////////////////////////////////////////////////////////////////
  if(!status) return -1;

  seg.clear();

  double min= 1e20;
  int smin=-1;
  for(int s=0; s<(int)seg_start.size(); s++){
    if(fabs(gps-seg_start[s])<min){
      min=fabs(gps-seg_start[s]);
      smin=s;
    }
    if(fabs(gps-seg_end[s])<min){
      min=fabs(gps-seg_end[s]);
      smin=s;
    }
  }

  if(smin==-1) return -1;

  seg.push_back(seg_start[smin]);
  seg.push_back(seg_end[smin]);

  return smin;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::ApplyPadding(const double pad_low, const double pad_up){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;
  if(!pad_low && !pad_up) return true; // nothing to do

  // tmp vectors  
  vector<double> seg_start_tmp;
  vector<double> seg_end_tmp;
  vector<int> seg_tag_tmp;

  // apply padding for each segment
  for(int s=0; s<(int)seg_start.size(); s++){
    seg_end_tmp.push_back(seg_end[s]+pad_up);
    seg_start_tmp.push_back(seg_start[s]+pad_low);
    seg_tag_tmp.push_back(seg_tag[s]);

    //remove empty/negative segments
    if(seg_end_tmp[seg_start_tmp.size()-1]<=seg_start_tmp[seg_end_tmp.size()-1]){
      seg_start_tmp.pop_back();
      seg_end_tmp.pop_back();
      seg_tag_tmp.pop_back();
    }
  }

  // switch to class vectors
  seg_end.clear();   seg_end = seg_end_tmp;
  seg_start.clear(); seg_start = seg_start_tmp;
  seg_tag.clear();   seg_tag = seg_tag_tmp;
  seg_end_tmp.clear();
  seg_start_tmp.clear();
  seg_tag_tmp.clear();

  // merge overlapping segments
  status*=MergeOverlaps();

  // set new time marks
  status*=SetMarks();

  // compute new livetime
  status*=ComputeLiveTime();

  return status;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::MakeSecond(void){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;

  // round starts and ends
  for(int s=0; s<(int)seg_start.size(); s++){
    seg_start[s] = floor(seg_start[s]);
    seg_end[s] = ceil(seg_end[s]);
  }

  // merge overlapping segments
  status*=MergeOverlaps();

  // set new time marks
  status*=SetMarks();

  // compute new livetime
  status*=ComputeLiveTime();

  return status;
}

/////////////////////////////////////////////////////////////////////////
double Segments::SmartShift(void){
/////////////////////////////////////////////////////////////////////////
  if(!status) return -1.0;
  if(!seg_start.size()) return 0.0;

  // tmp vectors  
  vector<double> seg_start_tmp;
  vector<double> seg_end_tmp;

  // smart-shift time segments
  for(int s=0; s<(int)seg_start.size()-1; s++){
    seg_start_tmp.push_back(seg_end[s]);
    seg_end_tmp.push_back(TMath::Min(2.0*seg_end[s]-seg_start[s],seg_start[s+1]));
  }
  seg_start_tmp.push_back(seg_end[seg_start.size()-1]);
  seg_end_tmp.push_back(2.0*seg_end[seg_start.size()-1]-seg_start[seg_start.size()-1]);
  

  // switch to class vectors
  seg_end.clear();   seg_end = seg_end_tmp;
  seg_start.clear(); seg_start = seg_start_tmp;
  seg_end_tmp.clear();
  seg_start_tmp.clear();

  // set new time marks
  status*=SetMarks();

  // compute new livetime
  status*=ComputeLiveTime();

  return livetime_tot;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::Intersect(Segments *Seg){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;
  if(Seg==NULL||!Seg->GetStatus()){
    cerr<<"Segments::Intersect: input segments are corrupted"<<endl;
    return false;
  }

  // new vectors  
  vector<double> seg_start_tmp;
  vector<double> seg_end_tmp;
  vector<int> seg_tag_tmp;

  int sstart=0;
  for(int s=0; s<Seg->GetNsegments(); s++){
    for(int ss=sstart; ss<(int)seg_start.size(); ss++){
      if(seg_end[ss]<=Seg->GetStart(s)){ 
	sstart=ss+1;
	continue; 
      }
      else if(seg_start[ss]<=Seg->GetStart(s)&&seg_end[ss]>=Seg->GetEnd(s)){ 
	sstart=ss; 
	seg_start_tmp.push_back(Seg->GetStart(s)); seg_end_tmp.push_back(Seg->GetEnd(s));
	seg_tag_tmp.push_back(TagProduct(seg_tag[ss],Seg->GetTag(s)));
	break; 
      }
      else if(seg_start[ss]<=Seg->GetStart(s)&&seg_end[ss]>Seg->GetStart(s)&&seg_end[ss]<=Seg->GetEnd(s)){
	sstart=ss+1; 
	seg_start_tmp.push_back(Seg->GetStart(s)); seg_end_tmp.push_back(seg_end[ss]);   
	seg_tag_tmp.push_back(TagProduct(seg_tag[ss],Seg->GetTag(s)));
      }
      else if(seg_start[ss]>=Seg->GetStart(s)&&seg_end[ss]<=Seg->GetEnd(s)){
	sstart=ss+1; 
	seg_start_tmp.push_back(seg_start[ss]); seg_end_tmp.push_back(seg_end[ss]);   
	seg_tag_tmp.push_back(TagProduct(seg_tag[ss],Seg->GetTag(s)));
      }
      else if(seg_start[ss]>=Seg->GetStart(s)&&seg_start[ss]<Seg->GetEnd(s)){
	sstart=ss; 
	seg_start_tmp.push_back(seg_start[ss]); seg_end_tmp.push_back(Seg->GetEnd(s));
	seg_tag_tmp.push_back(TagProduct(seg_tag[ss],Seg->GetTag(s)));
      }
      else{
	break;
      }
    }
  }
  
  // switch to class vectors
  seg_end.clear();   seg_end = seg_end_tmp;
  seg_start.clear(); seg_start = seg_start_tmp;
  seg_tag.clear();   seg_tag = seg_tag_tmp;

  // merge overlapping segments
  status*=MergeOverlaps();
      
  // set new time marks
  status*=SetMarks();
  
  // compute new livetime
  status*=ComputeLiveTime();
  
  return status;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::Intersect(const double gps_start, const double gps_end, const int tag){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;
  if(gps_start>=gps_end) return false;

  Segments *stmp = new Segments(gps_start,gps_end,tag);
  if(!stmp->GetStatus()){
    delete stmp;
    return false;
  }

  if(!Intersect(stmp)){
    delete stmp;
    return false;
  }
  
  return true;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::Reverse(void){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;
  
  // no segment
  if(seg_start.size()==0){
    Reset();
    AddSegment(700000000,2000000000,-1);
    return true;
  }

  double e = seg_start[0];
  double s = seg_end[(int)seg_start.size()-1];

  // erase the 1st element
  seg_start.erase(seg_start.begin());

  // erase the last element
  seg_end.pop_back();

  // swap vectors
  seg_start.swap(seg_end);

  // add first and last segment
  seg_start.insert(seg_start.begin(),700000000);
  seg_end.insert(seg_end.begin(),e);
  seg_start.push_back(s);
  seg_end.push_back(2000000000);

  // clear all tags
  seg_tag.clear();
  seg_tag.assign(seg_start.size(),-1);
  
  // set new time marks
  status*=SetMarks();
  
  // compute new livetime
  status*=ComputeLiveTime();
  
  return status;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::SetTag(const int s, const int tag){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;
  if(s<0||s>=(int)seg_start.size()){
    cerr<<"Segments::SetTag: there is no "<<s<<" segment"<<endl;
    return false;
  }
  seg_tag.erase(seg_tag.begin()+s); seg_tag.insert(seg_tag.begin()+s,TMath::Max(-1,tag));
  return true;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::SetTags(const int tag){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;
  seg_tag.clear();
  seg_tag.assign(seg_start.size(),TMath::Max(-1,tag));
  return true;
}

/////////////////////////////////////////////////////////////////////////
int Segments::GetTag(const int s){
/////////////////////////////////////////////////////////////////////////
  if(!status) return 0;
  if(s<0||s>=(int)seg_tag.size()){
    cerr<<"Segments::GetTag: there is no '"<<s<<"' segment"<<endl;
    return -1;
  }
  return seg_tag[s];
}
/////////////////////////////////////////////////////////////////////////
int Segments::GetNSegmentsWithTag(const int tag){
/////////////////////////////////////////////////////////////////////////
  if(!status) return 0;
  int sum=0;
  for(int s=0; s<(int)seg_tag.size(); s++) if(seg_tag[s]==tag) sum++;  
  return sum;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::MergeOverlaps(const int s_start){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;

  // nothing to do if there is only one segment
  if(seg_start.size()<2) return true;
  
  for(int s=TMath::Max(0,s_start-1); s<(int)seg_start.size()-1; s++){

    if(seg_start[s+1]>seg_end[s]) continue;// normal situation: (s+1) is after (s) and disjoint

    if(seg_start[s+1]>=seg_start[s]&&seg_end[s+1]<=seg_end[s]){// (s+1) is completely inside (s)
      seg_start.erase(seg_start.begin()+s+1);
      seg_end.erase(seg_end.begin()+s+1);

      // the tag logic is described in TagSum()
      if(seg_tag[s]>seg_tag[s+1]) seg_tag.erase(seg_tag.begin()+s+1);
      else                        seg_tag.erase(seg_tag.begin()+s);

      s--;// do not move and test next segment
    }

    else if(seg_start[s+1]>=seg_start[s]&&seg_end[s+1]>=seg_end[s]){// (s+1) overlaps (s)
      seg_end.erase(seg_end.begin()+s);
      seg_start.erase(seg_start.begin()+s+1);
 
      // the tag logic is described in TagSum()
      if(seg_tag[s]>seg_tag[s+1]) seg_tag.erase(seg_tag.begin()+s+1);
      else                            seg_tag.erase(seg_tag.begin()+s);
     
      s--;// do not move and test next segment
    }
    else{// should not happen
      cerr<<"Segments::MergeOverlaps: should never happen"<<endl;
      status=false;
      return false;
    }
  }

  return CheckConsistency();
}

/////////////////////////////////////////////////////////////////////////
bool Segments::Dump(int ncol){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;

  if(ncol==2){
    for(int s=0; s<(int)seg_start.size(); s++) cout<<seg_start[s]<<" "<<seg_end[s]<<endl;
  }
  else if(ncol==3){
    for(int s=0; s<(int)seg_start.size(); s++) cout<<s<<" "<<seg_start[s]<<" "<<seg_end[s]<<endl;
  }
  else if(ncol==4){
    for(int s=0; s<(int)seg_start.size(); s++) cout<<s<<" "<<seg_start[s]<<" "<<seg_end[s]<<" "<<seg_end[s]-seg_start[s]<<endl;
  }
  else if(ncol==5){
    for(int s=0; s<(int)seg_start.size(); s++) cout<<s<<" "<<seg_start[s]<<" "<<seg_end[s]<<" "<<seg_end[s]-seg_start[s]<<" "<<seg_tag[s]<<endl;
  }
  else if(ncol==1){
    for(int s=0; s<(int)seg_start.size(); s++) cout<<seg_end[s]-seg_start[s]<<endl;
  }
  else if(ncol==-1){
    for(int s=0; s<(int)seg_tag.size(); s++) cout<<seg_tag[s]<<endl;
  }
  else{
    cerr<<"Segments::Dump: the "<<ncol<<" columns format is not supported."<<endl;
    return false;
  }
  return true;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::Write(const string txtfilename, const int ncol){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;
    
  // open output file
  ofstream txtfile(txtfilename.c_str(), ofstream::out);
  txtfile.flags ( ios::fixed );
  
  if(ncol==2){
    for(int s=0; s<(int)seg_start.size(); s++) txtfile<<seg_start[s]<<" "<<seg_end[s]<<endl;
  }
  else if(ncol==3){
    for(int s=0; s<(int)seg_start.size(); s++) txtfile<<s<<" "<<seg_start[s]<<" "<<seg_end[s]<<endl;
  }
  else if(ncol==4){
    for(int s=0; s<(int)seg_start.size(); s++) txtfile<<s<<" "<<seg_start[s]<<" "<<seg_end[s]<<" "<<seg_end[s]-seg_start[s]<<endl;
  }
  else if(ncol==5){
    for(int s=0; s<(int)seg_start.size(); s++) txtfile<<s<<" "<<seg_start[s]<<" "<<seg_end[s]<<" "<<seg_end[s]-seg_start[s]<<" "<<seg_tag[s]<<endl;
  }
  else if(ncol==1){
    for(int s=0; s<(int)seg_start.size(); s++) txtfile<<seg_end[s]-seg_start[s]<<endl;
  }
  else if(ncol==-1){
    for(int s=0; s<(int)seg_tag.size(); s++) txtfile<<seg_tag[s]<<endl;
  }
  else{
    cerr<<"Segments::Write: the "<<ncol<<" columns format is not supported."<<endl;
    txtfile.close();
    return false;
  }

  txtfile.close();

  return true;
}

/////////////////////////////////////////////////////////////////////////
bool Segments::WriteROOT(const string rootfilename){
/////////////////////////////////////////////////////////////////////////
  if(!status) return false;

  TTree *outtree = GetTree();

  TFile outfile(rootfilename.c_str(),"RECREATE");
  if(outfile.IsZombie()){// not existing file
    cerr<<"Segments::WriteROOT: cannot open "+rootfilename<<endl;
    delete outtree;
    return false;
  }
  
  outfile.cd();
  outtree->Write();
  outfile.Close();

  delete outtree;
  return true;
}
